import thebe.core.database as Database
import thebe.core.run as Run
import thebe.core.html as Html
import thebe.core.data as data
import thebe.core.constants as Constants 
import os, time, logging, json

class fileController:
    def __init__(self, fileLoc, socketio = None,\
            globalScope = {}, localScope = {}, isIpynb = False):

        self.socketio = None
        self.fileLoc = fileLoc
        self.localScope = localScope
        self.globalScope = globalScope
        self.isActive = False

    def checkUpdate(socketio, fileLocation, connected=False, isIpynb=False):
        '''
        Combines isModified and update functions. 
        '''

        #Get file target information from database if it exists
        Cells, GlobalScope, LocalScope  = Database.getLedger(fileLocation)

        #If it's modified or if it's the first time it has run, update.
        if isModified(fileLocation) or not GlobalScope:
            time.sleep(1)
            update(socketio, fileLocation, GlobalScope, LocalScope, Cells, isIpynb)
            time.sleep(1)

        elif connected==True:
    #        html=Html.convertLedgerToHtml(Cells)
            socketio.emit('show all', Cells)

        else:
            time.sleep(1)

    #Run code and send code and outputs to client
    def update(socketio, fileLocation, GlobalScope, LocalScope, Cells, isIpynb):

        #logging.debug('Cells in the beginning\n---------------\n%s'%(Cells,))

        '''
        Get some variables from database
        '''
        executions = Database.getExecutions(fileLocation)

        '''
        Get target file
        '''
        fileContent=''
        with open(fileLocation, 'r') as file_content:
            fileContent=file_content.read()
        '''
        Look at the file to see if anything has changed
        in the data.
        If there is a change, return updated ledger, and a list
        of cells that need executing.
        '''
        Cells = data.update(Cells, fileContent)

        '''
        Send a list of the cells that will run to the
        client so it can show what is loading.
        '''
    #    socketio.emit('show loading', htmlAllCells)

        '''
        Run the newly changed cells and return their output.
        '''
        output=Run.runNewCells(Cells, GlobalScope, LocalScope)

        '''
        Send output to client
        '''
        #socketio.emit('show output', output)
        executions += 1
        logging.info('The number of code executions is %d' % executions)
    #    html=Html.convertLedgerToHtml(Cells)
        socketio.emit('show all', Cells)

        '''
        Update the database with the fresh code.
        '''
        Database.update(fileLocation, Cells, GlobalScope, LocalScope, executions)
        if isIpynb:
            updateIpynb(fileLocation, Cells)

    def updateIpynb(fileLocation, Cells):
        '''
        Write the new changes to the ipynb file.
        '''
        with open(fileLocation.split('.')[0]+'.ipynb', 'w') as f:
            ipynb = Constants.getIpynb()
            ipynb['cells'] = Cells
            json.dump(ipynb, f)

    def isModified(fileLocation, x=1):
        '''
        Return true if the target file has been modified in the past x amount of time
        '''

        lastModified=os.path.getmtime(fileLocation)
        timeSinceModified=int(time.time()-lastModified)

        if timeSinceModified<=x:
            return True
        else:
            return False




import sqlite3, logging, dill, os, sys, time
import thebe.core.constants as Constant
import thebe.core.logger as Logger
from datetime import datetime

logger = Logger.getLogger('database.log', __name__)

def dict_factory(cursor, row):
    d = {}
    for idx, col in enumerate(cursor.description):
        if col[0] == 'cells' or col[0] == 'local_scope' or col[0] == 'global_scope':
            d[col[0]] = dill.loads(row[idx])
        else:
            d[col[0]] = row[idx]
    return d

DATABASE = '%s/database.sqlite' % os.path.dirname(os.path.abspath(__file__))

def getLedger(target):
    '''
    Get the data objects, if they do not exist create them.
    '''
    conn = sqlite3.connect(DATABASE, check_same_thread = False)
    conn.row_factory = dict_factory
    c = conn.cursor()
    c.execute('SELECT * FROM ledger WHERE name=?', (target,))
    fetched=c.fetchone()
    conn.close()
    if bool(fetched):
        r=fetched
        logger.info('Fetched Cells:\n-------------------------------\n%s'%(r['cells'],))
        return r['cells'], r['global_scope'], r['local_scope']
    else:
        createLedger(target)
        time.sleep(.5)
        return getLedger(target)

def createLedger(target):
    '''
    '''
    conn = sqlite3.connect(DATABASE, check_same_thread = False)
    conn.row_factory = dict_factory
    c = conn.cursor()
    logger.info('Adding\t%s\t to db...' % target)
    now=datetime.now().strftime('%a, %B, %d, %y')
    c.execute('INSERT INTO ledger (name, last_edit, created, cells, global_scope, local_scope, is_active) \
            VALUES (?,?,?,?,?,?,?)', (target, now, now, dill.dumps([]), dill.dumps({}), dill.dumps({}), False))
    conn.commit()
    conn.close()

def getExecutions(target):
    '''
    '''
    conn = sqlite3.connect(DATABASE, check_same_thread = False)
    conn.row_factory = dict_factory
    c = conn.cursor()
    c.execute('SELECT executions FROM ledger WHERE name=?', (target,))
    executions = c.fetchone()['executions']
    conn.close()
    return executions

def getIsActive(target):
    '''
    '''
    conn = sqlite3.connect(DATABASE, check_same_thread = False)
    conn.row_factory = dict_factory
    c = conn.cursor()
    c.execute('SELECT is_active FROM ledger WHERE name=?', (target,))
    isActive = c.fetchone()['is_active']
    conn.close()
    return isActive

def setIsActive(target):
    '''
    '''
    conn = sqlite3.connect(DATABASE, check_same_thread = False)
    conn.row_factory = dict_factory
    c = conn.cursor()
    c.execute('UPDATE ledger SET is_active=? WHERE name=?', \
            (True, target))
    conn.commit()
    conn.close()

def setActive(target, active):
    conn = sqlite3.connect(DATABASE, check_same_thread = False)
    conn.row_factory = dict_factory
    c = conn.cursor()
    c.execute('UPDATE ledger SET is_active=? WHERE name=?', (active, target))
    conn.commit()
    conn.close()

def update(target, cells, globalScope, localScope, executions):
    '''
    '''
    conn = sqlite3.connect(DATABASE, check_same_thread = False)
    conn.row_factory = dict_factory
    c = conn.cursor()
    localdump = {}
    try:
        localdump = dill.dumps(localScope)
    except AttributeError:
        logger.debug('Dill pickling the local scope, yields an error')
    logger.info('Before updating ledger with new cells:\n-------------------------------\n%s'%(cells,))
    c.execute('UPDATE ledger SET cells=?, global_scope=?, local_scope=?, executions=? WHERE name=?', \
            (dill.dumps(cells), dill.dumps(globalScope), dill.dumps(localdump), executions, target))
    conn.commit()
    conn.close()



from itertools import zip_longest
import time, sys, datetime, glob, re, sys, time, os, copy
from hashlib import md5
from io import StringIO
from subprocess import Popen, PIPE
from flask import url_for
from pygments import highlight
from pygments.lexers import BashLexer, PythonLexer
from pygments.formatters import HtmlFormatter
from flask_socketio import emit, SocketIO
import python_live.core.constants as Constant

def addSavePlot(fileString, myDir, plotCount):
    fileArrayWithPlot=[]
    savefigText="plt.savefig('"+myDir+"/static/plot"+str(plotCount)+".png')"
    match = re.compile(r"###p")
    items = re.findall(match, fileString)
    for item in items:
        savefigText="plt.savefig('"+myDir+"/static/plot"+str(plotCount)+".png')"
        fileString=fileString.replace(item, savefigText)
    return fileString 
def getPlotData(globalScope, localScope):
    code=Constant.GetPlot
    redirected_output=sys.stdout=StringIO()
    redirected_error=sys.stderr=StringIO()
    stdout=''
    stderr=''
    try:
        sys.path.append(os.getcwd())
        exec(code, globalScope, localScope)
        stdout=redirected_output.getvalue()
        stderr=''
    except Exception as e:
        stdout=redirected_output.getvalue()
        stderr=str(e)
    finally:
        sys.path.pop()
        sys.stdout=sys.__stdout__
        sys.stderr=sys.__stderr__
    if stdout==Constant.EmptyGraph:
        stdout=''
    return stdout
def updateChanged(cellsList):
    for cell in cellsList:
        cell['changed']=False
def updateLedgerPop(oldAllCells, fileString, ledger, myDir):
    allCellsList=[]
    newLedger=[]
    newCellsToRun=[]
    cellDelimiter='####\n'
    for cellCount, cell in enumerate(list(filter(None, fileString.split(cellDelimiter)))):
        cellHash=md5(cell.encode()).hexdigest()
        newLedger.append(cellHash)
        isChanged=False
        stdout=''
        stderr=''
        if cellHash not in ledger:
            isChanged=True
        for oldCell in oldAllCells:
#            print(oldCell['stdout'])
#            print(oldCell['hash'])
            if oldCell['hash']==cellHash:
                stdout=oldCell['stdout']
                stdtr=oldCell['stderr']
        allCellsList.append({'cellCount':str(cellCount), 'hash':cellHash, 'code':cell, 'plot':'', 'datetime': time.strftime("%x %X", time.gmtime()), 'changed': isChanged, 'stdout':stdout, 'stderr':stderr, 'image/png':''})
    newCellsToRun=getNewCellsToRun(ledger, allCellsList)
    return newLedger, allCellsList
def getNewCellsToRun(ledger, allCellsList):
    newCellsToRun=[]
    for currentCell, ledgerCell in zip_longest(allCellsList, ledger, fillvalue=None):
        if currentCell['hash']!=ledgerCell:
            currentCell['changed']=True
        newCellsToRun.append(currentCell)
    return newCellsToRun
def convertLedgerToHtml(cellList, part='all'):
    # Return a deep copy of cellList with code replaced with html-ized code
    tempCells=copy.deepcopy(cellList)
    for cell in tempCells:
        if part=='all':
            cell['code']=highlight(cell['code'], PythonLexer(), HtmlFormatter())
            if cell['image/png']:
                cell['image/png']='<img src="data:image/png;base64, '+cell['image/png']+'" />'
            if cell['stdout']:
                cell['stdout']=highlight(cell['stdout'], BashLexer(), HtmlFormatter())
            if cell['stderr']:
                cell['stderr']=highlight(cell['stderr'], BashLexer(), HtmlFormatter())
        if part=='output':
            if cell['image/png']:
                cell['image/png']='<img src="data:image/png;base64, '+cell['image/png']+'" />'
            if cell['stdout']:
                cell['stdout']=highlight(cell['stdout'], BashLexer(), HtmlFormatter())
            if cell['stderr']:
                cell['stderr']=highlight(cell['stderr'], BashLexer(), HtmlFormatter())

    return tempCells 
def runNewCells(cellsToRun, ledger, globalScope, localScope, myDir):
    cellOutput=[]
    for cellCount, cell in enumerate(cellsToRun):
        if cell['changed']==True:
            stdout, stderr, plotData=runWithExec(cell['code'], globalScope, localScope)
            # Keep the master list updated
            cellsToRun[cellCount]['stdout']=stdout
            cellsToRun[cellCount]['stderr']=stderr
            cellsToRun[cellCount]['image/png']=plotData
            cellOutput.append({'stdout':stdout, 'stderr':stderr, 'image/png':plotData})
        else:
            cellOutput.append('')
    return cellOutput
def runWithExec(cellCode, globalScope, localScope):
    redirected_output=sys.stdout=StringIO()
    redirected_error=sys.stderr=StringIO()
    stdout=''
    stderr=''
    try:
        sys.path.append(os.getcwd())
        exec(cellCode, globalScope, localScope)
        stdout=redirected_output.getvalue()
        stderr=''
    except Exception as e:
        stdout=redirected_output.getvalue()
        stderr=str(e)
    finally:
        sys.path.pop()
        sys.stdout=sys.__stdout__
        sys.stderr=sys.__stderr__
    plotData=getPlotData(globalScope, localScope)
    return stdout, stderr, plotData



import os, json, logging, sys
import thebe.core.data as data 

def setup(targetName):
    '''
    Checks if the command line argument file is ipynb format,
    if it is, write a new file of the same prefix to the same
    directory, containing only the source files in thebe format.
    Return that new file name and a true flag.

    If .py file extension, return the same file
    name that was inputted and a false flag.
    '''
    if isIpynb(targetName):
        ipynbFileContent = loadIpynb(targetName)
        tempName = write_temp(data.toThebe(ipynbFileContent), targetName)
        return tempName, True
    else:
        return targetName, False

        
def isIpynb(targetLocation):
    '''
    Return the target's file extension. 
    If input is incorrect, explain, and quit the application.
    '''
    if os.path.isfile(targetLocation):
        try:
            return test_extension(targetLocation) == 'ipynb'
        except ValueError:
            print('Please use a valid file extension. (.ipynb or .py)')
            sys.exit()
    else:
        print('Thebe only works with files, not directories. Please try again with a file. (.ipynb or .py)')
        sys.exit()

def test_extension(targetLocation):
    '''
    '''

    targetExtension=targetLocation.split('.')[-1]
    if targetExtension=='ipynb':
        return 'ipynb'
    elif targetExtension=='py':
        return 'py'
    else:
        print('Please use a valid file extension. (.ipynb or .py)')
        sys.exit()

def loadIpynb(fileName):
    '''
    Return the ipynb file as a dictionary.
    '''

    data = {}
    with open(fileName) as ipynb_data:
        data = json.load(ipynb_data)
    return data

def write_temp(initialData, fileName):
    '''
    Create a temporary file and load it with our initial data from the ipynb file.
    Open vim in our temporary file.
    '''
    filePrefix = getPrefix(fileName)
    tempFileName = '%s.py'%(filePrefix)

    try:
        with open(tempFileName, 'w') as f:
            f.write(initialData)
    except KeyboardInterrupt:
        removeTemp()

    return tempFileName

def removeTemp(tempFileName):
    '''
    '''
    os.remove(tempFileName)

def getPrefix(fileName):
    '''
    Get the file prefix
    '''

    # Doing this in a weird way to handle for situations where there are
    # dots other than the one signifying extension.
    splitName = fileName.split('.')
    splitName.pop()
    return '.'.join(splitName)





from itertools import zip_longest
import time, sys, datetime, glob, re, sys, time, os, copy, logging
from hashlib import md5
from io import StringIO
from subprocess import Popen, PIPE
from flask import url_for
from pygments import highlight
from pygments.lexers import BashLexer, PythonLexer
from pygments.formatters import HtmlFormatter
from flask_socketio import emit, SocketIO
import thebe.core.constants as Constants
import thebe.core.logger as Logger

logger = Logger.getLogger('data.log', __name__)

def update(oldCellList, fileContent, runAll=False):
    '''
    Determine which cell has changed, since the file has changed.
    Return a list of cells, with updated sources, and the changed
    variable set.
    '''

    cellList = []
    sourceList = getSourceList(oldCellList)

    # Ignore code that comes before the first delimiter
    ignoreFirst = False
    if not fileContent.startswith(Constants.CellDelimiter):
        logger.info('Ignoring first...')
        logger.info('---------------\n%s\n-------------'%(fileContent[0:10],))
        ignoreFirst = True

    for cellCount, source in enumerate(fileContent.split(Constants.CellDelimiter)):

        # Ignore the first source if a cell delimiter
        # does not preceed it
        if ignoreFirst:
            ignoreFirst = False
            continue

        #Split source by line
        source = source.splitlines(True)

        if validSource(source):
            # Get copy of cell initially populated
            cell = setChanged(oldCellList, sourceList, source)

            # If outputs are empty, populate outputs
            # with an initial empty output so we 
            # don't get empty list errors
            if not cell['outputs']:
                cell['outputs'] = [Constants.getExecuteOutput()]

            # This activates if the user clicks the run all button
            if runAll:
                cell['changed'] = True

            # Set execution counter if it exists
            try:
                cell['execution_count'] = oldCellList[cellCount]['execution_count'] 
            except IndexError:
                pass

            cellList.append(cell)

    return cellList

def getSourceList(cellList):
    '''
    Form the hashes from the cell list into a list
    '''
    return [cell['source'] for cell in cellList]

def toThebe(ipynb):
    '''
    Take in a ipynb dictionary, and returns a string in thebe format.
    (Cell sources to delimited by our Constants.CellDelimiter)
    '''

    output = ''
    for cell in ipynb['cells']:
        if cell['cell_type'] == 'markdown':
            output = ''. join(\
                    (output, \
                    (Constants.CellDelimiter + 'm\n' + \
                    ''.join(cell['source']))\
                    ))
        else:
            output = ''. join(\
                    (output, \
                    (Constants.CellDelimiter + '\n' + \
                    ''.join(cell['source']))\
                    ))
#            logger.info('ipynb conversion output:\
#                    \n-------------------------------\%s'%\
#                    (output,))
    return output
    

def setChanged(oldCellList, sourceList, source):
    '''
    If the source preexists, set new cell to old cell
    If not, set changed, and last changed time.
    '''

    cell=copy.deepcopy(Constants.Cell)

    # Detect if cell is Markdown
    if source[0] == 'm\n':
        # Remove the markdown identifier
        source.pop(0)
        # Set cell as markdown
        cell['cell_type'] = 'markdown'
    else:
        # Remove the new line after the delimiter
        source.pop(0)

    #Set sourceCode
    cell['source'] = source

    try:
        x = sourceList.index(source)
        cell = oldCellList[x]

    except ValueError:
        cell['execution_count']
        cell['changed']=True
        cell['last_changed']=time.strftime("%x %X", time.gmtime())

    return cell

def hashSource(source):
    '''
    Hash the string of code
    '''
    return md5(source.encode()).hexdigest()

def validSource(source):
    '''
    Return false if source list is all Just new lines
    '''
    for s in source:
        if s != '\n':
            return True
    return False




import sys, logging
from io import StringIO

def reroute(func, *args):
    '''
    Reroute standard output during an individual function,
    and return the outputs.
    '''
    stdout = sys.stdout = StringIO()
    stderr = sys.stderr = StringIO()
    func(*args)
    sys.stdout = sys.__stdout__
    sys.stderr = sys.__stderr__

    return stdout.getvalue(), stderr.getvalue()

class OutputController:

    def __init__(self):
        self.stdout = []
        self.stderr = []

    def open(self):
        '''
        Append an object to our outputs list that output
        will now be piped to.
        '''
        sys.stdout = StringIO()
        sys.stderr = StringIO()

        self.stdout.append(sys.stdout)
        self.stderr.append(sys.stderr)
        

    def close(self):
        '''
        Pop from the outputs list, return the outputs as strings.
        Begin piping to the next object in the outputs list.
        '''
        tstdout = ''
        tstderr = ''
        if self.stdout and self.stderr:
#            logging.info([x for x in self.stdout])
            #If the stack is not empty, save the outputs to variables that will be returned at the end of this function.
            tstdout = self.stdout.pop().getvalue()
            tstderr = self.stderr.pop().getvalue()
            logging.debug('Standard output:\t%s\tStandard error:\t%s'%(tstdout, tstderr))
            
            if self.stdout and self.stderr:
                #If the stack is not empty, then pipe output to the next object in the stack.
                sys.stdout = self.stdout[-1]
                sys.stderr = self.stderr[-1]
            else:
                #If the stack is empty, reset output to default.
                sys.stdout = sys.__stdout__
                sys.stderr = sys.__stderr__
        else:
            logging.info('The output stack is empty, please open a new output to add to the stack.')
        return tstdout, tstderr 
    
    def pclose(self):
        '''
        Run the close function, logging.info its output
        '''
        logging.info('stdout\n------------\n%s\nstderr\n----------\n%s'%self.close())



import time, sys, datetime, glob, re, sys, time, os, copy, logging, threading
logging.StreamHandler(stream=None)

def getLogger(fileLoc, name):
    logger = logging.getLogger(name)

    file_handler = logging.FileHandler('%s/logs/%s'%(os.path.dirname(__file__), fileLoc), mode='w')

    formatter = logging.Formatter('%(message)s')
    streamHandler = logging.StreamHandler()
    streamHandler.setFormatter(formatter)

    logger.setLevel(logging.DEBUG)

    logger.addHandler(file_handler)
    logger.addHandler(streamHandler)
    logging.StreamHandler(stream=None)

    return logger




import os,sys, tempfile
from io import StringIO
#from thebe.core.output import outputController
from multiprocessing import Process
from subprocess import call, check_output

class FileManager:
    def __init__(self, target_name):
        self.temp_name = ''
        self.target_name = target_name

        target_ext = self.test_file(self.target_name)

        if target_ext == 'ipynb':
            self.ipynb_name = ipynb_name
            ipynb_content = self.load_ipynb(target_name)
            self.target_name = self.write_temp(ipynb_content)


def test_file(targetLocation):
    '''
    Return the relevant extension. 
    If input is incorrect, explain, and quit the application.
    '''
    if os.path.isfile(targetLocation):
        try:
            return test_extension(targetLocation)
        except ValueError:
            logging.info('Please use a valid file extension. (.ipynb or .py)')
            sys.exit()
    else:
        logging.info('Thebe only works with files, not directories. Please try again with a file. (.ipynb or .py)')
        sys.exit()

def test_extension(targetLocation):
    '''
    '''

    targetExtension=targetLocation.split('.')[1]
    if targetExtension=='ipynb':
        return 'ipynb'
    elif targetExtension=='py':
        return 'py'
    else:
        logging.info('Please use a valid file extension. (.ipynb or .py)')
        sys.exit()

def load_ipynb(targetLocation):
    '''
    Return the ipynb file as a dictionary.
    '''

    data = {}
    with open(targetLocation) as ipynb_data:
        data = json.load(ipynb_data)
    return data
    def open(self):
        '''
        '''
        print('This is the location of the temporary file:\t%s'%(self.temp_loc))

        def callVim():
            # Open the file with the text editor
    #        outputController.open()
            so = sys.stdout = StringIO()
            EDITOR = os.environ.get('EDITOR','vim')
            call([EDITOR, self.temp_loc])
    #        outputController.close()

        try:
            print('Starting vim process...')
            vim = Process(target = callVim)
            vim.start()

        except KeyboardInterrupt:
            print("Terminating vim server.")
            vim.terminate()
            vim.join()
            print("Terminated flask server.")




import datetime, glob, re, sys, time, os, copy, logging
from pygments import highlight
from pygments.lexers import BashLexer, PythonLexer, MarkdownLexer
from pygments.formatters import HtmlFormatter
import pypandoc

def convert(cellList):
    '''
    Return a deep copy of cellList with code replaced with html-ized code
    '''

    # Deep copy cells so the original is not converted to HTML
    tempCells=copy.deepcopy(cellList)

    for cell in tempCells:
        # Preprocessing a code cell
        if cell['cell_type'] == 'code':
            # Highlight the Python syntax
            cell['source'] = \
                    [highlight(source, PythonLexer(), HtmlFormatter()) for source in cell['source']] 
            
            # Highlight standard output and error
            for output in cell['outputs']:
                # Highlight the standard output
                if output['output_type'] == 'execute_result':
                    output['data']['text/plain'] = \
                            [highlight(text, BashLexer(), HtmlFormatter()) \
                            for text in output['data']['text/plain']]
                # Highlight the error
                if output['output_type'] == 'error':
                    output['traceback'] = \
                            [highlight(text, BashLexer(), HtmlFormatter()) \
                            for text in output['traceback']]

        # Preprocessing a markdown cell
        if cell['cell_type'] == 'markdown':

            # Flatten the list so multi line latex delimiters 
            # are not separated by HTML elements as this would
            # break MathJax.
            cell['source'] = ''.join(cell['source'])

            # Remove any html that could interupt 
            # markdown conversion
            clean = re.compile('<.*>.*</.*>')
            cell['source'] = re.sub(clean, '', cell['source']) 

            # Convert the markdown

            # These arguments are used to let pypandoc
            # know to ignore latex
            pdoc_args = ['--standalone', '--mathjax'] 
            # Convert from markdown to HTML
            cell['source'] = \
                    pypandoc.convert_text(cell['source'], \
                    'html', format = 'md',\
                    extra_args = pdoc_args)

    return tempCells 

def convertText(text, ttype = 'bash'):
    return highlight(text, BashLexer(), HtmlFormatter()) 

def output(output):
    temp_output=copy.deepcopy(output)
    for cell in temp_output:
        if cell:
            if cell['image/png']:
                cell['image/png']='<img src="data:image/png;base64, '+cell['image/png']+'" />'
            if cell['stdout']:
                cell['stdout']=highlight(cell['stdout'], BashLexer(), HtmlFormatter())
            if cell['stderr']:
                cell['stderr']=highlight(cell['stderr'], BashLexer(), HtmlFormatter())
    return temp_output




#!/usr/bin/env python3

from jupyter_client.manager import start_new_kernel
from pygments.formatters import HtmlFormatter
from flask import Flask, render_template, Response, g
from flask_socketio import SocketIO, emit
from multiprocessing import Process
import thebe.core.update as Update
import thebe.core.args as args
import thebe.core.logger as Logger
from thebe.core.jupyter_wrapper import jupyter_client_wrapper
import tempfile, time, os, sys, webbrowser, logging, logging.config, json

port = args.getPort()
input_filename = args.getFile()

#Initialize flask
app = Flask(__name__)
app.config['SEND_FILE_MAX_AGE_DEFAULT'] = 0
app.config['SECRET_KEY'] = 'secret!'
socketio = SocketIO(app, async_handlers = False)

# Set up the execution engine
jupyter = jupyter_client_wrapper(input_filename)
jupyter.setSocket(socketio)

#Configure logging
root = logging.getLogger()
handler = logging.StreamHandler(None)
root.addHandler(handler)
logger = Logger.getLogger('sockets.log', __name__)
logging.StreamHandler(stream=None)
if not args.getDebug():
    logging.disable(logging.INFO)

'''
Set some headers and get and send css for all of the HtmlFormatter components.
'''
@app.route('/')
def home():
    css=HtmlFormatter().get_style_defs('.highlight')
    response=Response(render_template('main.html', css=css))
    response.headers["Cache-Control"] = "no-cache, no-store, must-revalidate" # HTTP 1.1.
    response.headers["Pragma"] = "no-cache" # HTTP 1.0.
    response.headers["Expires"] = "0" # Proxies
    return response

'''
Connect and disconnect events.
'''
@socketio.on('connect')
def connect():
    logger.info('Connected to client')
    #Show
    jupyter.execute(update = 'connected')
    #Start pinging
    socketio.emit('ping client')

@socketio.on('disconnect')
def disconnect():
    logger.info('Client disconnected')

'''
Ping back and forth from client to server.
Checks whether or not the file has been saved and running it when changed.
'''
@socketio.on('check if saved')
def check():
    jupyter.execute(update = 'changed')
    socketio.emit('ping client')

'''
'''
@socketio.on('run_all')
def run_all():
    '''
    Restart the kernel, and clear saved code.
    '''
    # Restart the kernel after active is set
    jupyter.execute(update = 'all')
@socketio.on('run cell')
def run_all(index):
    '''
    Restart the kernel, and clear saved code.
    '''
    # Restart the kernel after active is set
    jupyter.execute(update = [index])

'''
Run flask and socketio.
'''
def main():
    try:
        print('Running Thebe server on port %s...\nTo access Thebe, go to: localhost:%s\n\nTo quit, press Ctrl-c twice...'%(port,port))
        socketio.run(app, port=port, debug=False)
    except KeyboardInterrupt:
        print("Shutdown requested...exiting")
        socketio.stop()



from jupyter_client.manager import start_new_kernel
import thebe.core.file as File
import thebe.core.constants as Constant
import thebe.core.output as output 
import thebe.core.logger as Logger
import thebe.core.database as Database
import thebe.core.html as Html
from pygments import highlight
from pygments.lexers import BashLexer, PythonLexer, MarkdownLexer
from pygments.formatters import HtmlFormatter
import pypandoc, time, sys, datetime, glob, re, sys, time, os, copy, logging, threading, queue, json

class jupyter_client_wrapper:
    def __init__(self, fileName):
        # Setting up self.loggers
        self.lUpdate = Logger.getLogger('update.log', 'update')
        self.logger = Logger.getLogger('main.log', 'main')
        self.mess_logger = Logger.getLogger('mess.log', 'mess')
        self.status_logger = Logger.getLogger('status.log', 'status')
        self.execute_logger = Logger.getLogger('execute.log', 'execute')


        self.kernel_manager, self.jupyter_client = start_new_kernel()
        # Determine what kind of file working with
        self.fileLocation, is_ipynb = File.setup(fileName)

        # Initialize some variables
#        self.isActive = False
        self.executionThread = None
        self.Cells = []
        self.executions = 0

        # Initialize local and global scope for old
        # execution engine
        self.LocalScope = {}
        self.GlobalScope = {}

    '''
    -------------------------------
    Setter functions
    -------------------------------
    '''
    def setSocket(self, socketio):
        self.socketio = socketio

    '''
    -------------------------------
    Execution pre-preprocessing
    -------------------------------
    '''

    def execute(self, update = 'changed'):
        #self.Cells = Database.getCells(self.fileLocation)
        if self._isActive():
            self.socketio.emit('flash')
        else:
            if update == 'changed':
                self._execute_changed()

            elif update == 'all':
                self._execute_all()

            elif update == 'connected':
                self._execute_connected()

            elif type(update) == list:
                self._execute_cell(update)

            else:
                pass

        time.sleep(.5)
    '''
    -------------------------------
    Execution pre-preprocessing - Helpers
    -------------------------------
    '''
    def _execute_cell(self, index):
#        self.isActive = True
        self._execute(update = index)
#        self.isActive = False

    def _execute_changed(self):
        if self._isModified():
            self.status_logger.info('Execute changed...')
#            self.isActive = True
            self._execute(update = 'changed')
#            self.isActive = False
            self.status_logger.info('Finished...')

    def _execute_all(self):
        self.status_logger.info('Execute all...')
#        self.isActive = True
        self.Cells = []
        self._execute(update = 'all')
#        self.isActive = False
        self.status_logger.info('Finished...')

    def _execute_connected(self):
        if not self.Cells:
            self.status_logger.info('Execute connected...')
#            self.isActive = True
            self._execute(update = 'connected')
#            self.isActive = False
            time.sleep(2)
        else: 
            self.socketio.emit('show all', self._convert())

    def _restart_kernel(self):
        self.status_logger.info('Restarting kernel...')
        kernel_manager.restart_kernel()
        while True:
            try:
                io_msg_content = self.jupyter_client.get_iopub_msg(timeout=3)['content']
                time.sleep(.01)

            except queue.Empty:
                break

    def _isModified(self, x=.3):
        '''
        Return true if the target file has been modified in the past x amount of time
        '''

        lastModified=os.path.getmtime(self.fileLocation)
        timeSinceModified=int(time.time()-lastModified)
        if timeSinceModified<=x:
            return True
        else:
            return False

    '''
    -------------------------------
    Execution - Main - Helpers
    -------------------------------
    '''

    def _execute(self, update = 'changed'):
        '''
        Execute cells based on 'update' value
        '''
        self.status_logger.info('Updating...')
        self._update(self._getFileContent(), update)
        self.status_logger.info('Showing...')
        self.socketio.emit('show all', self._convert())
        '''
        '''
        self.status_logger.info('Executing...')
        self.executionThread = threading.Thread(target = self._executeThread)
        self.executionThread.daemon = True
        self.executionThread.start()

    def _update(self, fileContent, update):
        '''
        Update the 'Cells' variable with new data from
        the Thebe file
        '''

        # 'cells' is for the new cells
        cells = []

        # Ignore code that comes before the first delimiter
        ignoreFirst = self._ignoreFirst(fileContent)
        self.lUpdate.info('---------------')
        self.lUpdate.info('ignoreFirst: %s'%(ignoreFirst,))
        # True cell count, used because sourceCount
        # can be unreliable
        cellCount = 0

        for sourceCount, source in enumerate(fileContent.split(Constant.CellDelimiter)):

            self.lUpdate.info('---------------')
            self.lUpdate.info('Cell count: %s'%(cellCount,))
            # Ignore the first source if a cell delimiter
            # does not preceed it
            if ignoreFirst:
                ignoreFirst = False
                continue

            #Split source by line
            source = source.splitlines(True)
            self.lUpdate.info('Source length: %s'%(len(source),))

            if self._validSource(source):
                # Get copy of cell initially populated
                cell = copy.deepcopy(Constant.Cell)

                cell = self._setMarkdown(source, cell)
                self.lUpdate.info('Changed: %s'%(cell['changed'],))

                # Handle front end buttons
                if update == 'all':
                    cell = self._setChanged(source, cell, force_change = True)
                elif type(update) == list and cellCount in update:
                        cell = self._setChanged(source, cell, force_change = True)
                else:
                    cell = self._setChanged(source, cell)
                self.lUpdate.info('Changed: %s'%(cell['changed'],))

                # Set execution counter if it exists previously
                try:
                    cell['execution_count'] = self.Cells[cellCount]['execution_count'] 
                except IndexError:
                    pass
                cells.append(cell)
                cellCount  += 1

        self.Cells = cells
        self.status_logger.info('Changed cells: %s'%([x['changed'] for x in self.Cells],))

    def _executeThread(self):
        '''
        Run the newly changed cells and return their output.
        '''
        self.status_logger.info('Inside execute thread...')
        self.Cells = self._runNewCells()
        self.executions += 1

        '''
        Update the database with the fresh code.
        And outputs.
        '''
#        Database.update(self.fileLocation, self.Cells, GlobalScope, LocalScope, executions)
        self._updateIpynb()
        self.status_logger.info('Execution thread finished...')
    '''
    -------------------------------
    Actual execution
    -------------------------------
    '''
    def _runNewCells(self):
        '''
        Run each changed cell, returning the output.
        '''

        # Append cells containing updated output to this
        newCells = []

        # Toggle to true when the users code produces an
        # error so code execution can stop
        kill = False

        for cellCount, cell in enumerate(self.Cells):

            # Run changed code if it is not markdown
            # and no prior cell has triggered an error
            if cell['changed']:

                self.socketio.emit('message', 'Running cell #%s...'%(cellCount))
                self.socketio.emit('loading', cellCount)
                cell['changed'] = False
                if cell['cell_type'] != 'markdown' and not kill:
                    self.logger.info('\n------------------------\nRunning cell #%s\nIn directory: %s\nWith code:\n%s\n-------------------------------'%(cellCount, os.getcwd(), cell['source']))

                    # Execute the code from the cell, stream 
                    # outputs using socketio, and return output
                    outputs = self._jExecute(cell['source'])

                    # Prevent subsequent execution of code
                    # if in error was found
                    if self._hasError(outputs):
                        kill = True

                    # Add output data to cell
                    cell['outputs'] = outputs

                    # How does ipython do this?
                    cell['execution_count'] = cell['execution_count'] + 1

            # Append run cell the new cell list
            newCells.append(cell)

        # Stop the front end loading
        self.socketio.emit('stop loading')
        self.status_logger.info('End of runAllCells...')
        return newCells

    def _jExecute(self, code):
        '''
        '''

        self.status_logger.info('Inside jExecute thread...')
        code = ''.join(code)

        # Execute the code
        msg_id = self.jupyter_client.execute(code)

        # Get the execution status
        # When the execution state is "idle" it is complete
        t = True
        try:
            io_msg_content = self.jupyter_client.get_iopub_msg(timeout=1)['content']

        except queue.Empty:
            t = False

        self.status_logger.info('After first message in jExecute...')

        # Initialize the temp variable
        temp = {}

        # Initialize outputs
        outputs = []

        # Continue polling for execution to complete
        # which is indicated by having an execution state of "idle"
        while t:
            # Save the last message content. This will hold the solution.
            # The next one has the idle execution state indicating the execution
            # is complete, but not the stdout output
            temp = io_msg_content
            self.execute_logger.info(temp)
            # Check the message for various possibilities
            if 'data' in temp: # Indicates completed operation
                if 'image/png' in temp['data']:
                    plotData =  temp['data']['image/png']
                    output = self._getPlotOutput(plotData)
                    outputs.append(output)

                    self.socketio.emit('output', output)

            if 'name' in temp and temp['name'] == "stdout": # indicates output
                # Create output for server use
                output = self._getStdOut(temp['text'])
                outputs.append(output)

                # Send HTML output for immediate front end use
                htmlOutput = copy.deepcopy(output)
                htmlOutput['data']['text/plain'] = [Html.convertText(text, ttype = 'bash') for text in htmlOutput['data']['text/plain']]
                self.socketio.emit('output', htmlOutput)

            if 'evalue' in temp: # Indicates error

                # Create output for server use
                output = self._getErr(temp['evalue'])
                outputs.append(output)

                # Send HTML output for immediate front end use
                htmlOutput = copy.deepcopy(output)
                htmlOutput['evalue'] = [Html.convertText(text, ttype = 'bash') for text in htmlOutput['evalue']]
                self.socketio.emit('output', htmlOutput)
                
                # If there is an error than it is pointless 
                # to keep on running code
                break

            # Poll the message
            try:
                io_msg_content = self.jupyter_client.get_iopub_msg(timeout=1)['content']
                time.sleep(.1)

                if 'execution_state' in io_msg_content and io_msg_content['execution_state'] == 'idle':
                    break

            except queue.Empty:
                break

        self.status_logger.info('End of jExecute thread...')
        return outputs

    '''
    -------------------------------
    Execution helper functions
    -------------------------------
    '''
    def _hasError(self, outputs):
        '''
        If an error cell exists in outputs
        return true
        '''
        for output in outputs:
            if 'evalue' in output:
                return True
        return False

    def _fillPlot(cell, plot):
        '''
        If an image exists in the plot variable, create and return a plot cell.
        '''
        if plot:
            output = Constant.getDisplayOutput()
            output['data']['image/png'] = plot
            cell['outputs'].append(output)
        return cell

    def _getPlotOutput(self, plot):
        '''
        '''
        output = Constant.getDisplayOutput()
        output['data']['image/png'] = plot
        return output

    def _getStdOut(self, stdOut):
        output = Constant.getExecuteOutput()
        output['data']['text/plain'] = stdOut.splitlines(True)
        return output

    def _getErr(self, err):
        output = Constant.getErrorOutput()
        output['evalue'] = err.splitlines(True)
        return output
    '''
    -------------------------------
    Preprocessing helper functions
    -------------------------------
    '''

    def _getFileContent(self):
        fileContent = ''
        with open(self.fileLocation, 'r') as file_content:
            fileContent = file_content.read()
        return fileContent

    def _update_all(self):
        # 'cells' is for the new cells
        cells = []


    def _ignoreFirst(self, fileContent):
        # Ignore code that comes before the first delimiter
        if not fileContent.startswith(Constant.CellDelimiter):
            self.logger.info('Ignoring first...')
            return True
        else:
            return False

    def _getSourceList(self):
        '''
        Form the hashes from the cell list into a list
        '''
        return [cell['source'] for cell in self.Cells]

    def _toThebe(self, ipynb):
        '''
        Take in a ipynb dictionary, and returns a string in thebe format.
        (Cell sources to delimited by our Constants.CellDelimiter)
        '''

        output = ''
        for cell in ipynb['cells']:
            if cell['cell_type'] == 'markdown':
                output = ''. join(\
                        (output, \
                        (Constant.CellDelimiter + 'm\n' + \
                        ''.join(cell['source']))\
                        ))
            else:
                output = ''. join(\
                        (output, \
                        (Constant.CellDelimiter + '\n' + \
                        ''.join(cell['source']))\
                        ))
        return output
        

    def _setChanged(self, source, cell, force_change = False):
        '''
        Detect if a cell has been changed
        '''

        if force_change:
            cell['changed'] = True
            cell['last_changed'] = time.strftime("%x %X", time.gmtime())
            cell['outputs'] = []
            
        else:
            try:
                x = self._getSourceList().index(source)
                cell = self.Cells[x]

            except ValueError:
                cell['changed'] = True
                cell['last_changed'] = time.strftime("%x %X", time.gmtime())

        return cell

    def _setMarkdown(self, source, cell):
        '''
        Determine if a cell is marked down
        '''

        # Detect if cell is Markdown
        if source[0] == 'm\n':
            # Set cell as markdown
            cell['cell_type'] = 'markdown'

        # Remove the new line after the delimiter
        source.pop(0)

        #Set sourceCode
        cell['source'] = source

        return cell

    def _validSource(self, source):
        '''
        Return false if source list is all Just new lines
        '''
        for s in source:
            if s != '\n':
                return True
        return False

    def _updateIpynb(self):
        '''
        Write the new changes to the ipynb file.
        '''
        self.status_logger.info('Updating .ipynb...')
        cCells = copy.deepcopy(self.Cells)

        # Remove extra attributes created by thebe.
        self._sanitize(cCells)
        

        # Save cells into a ".ipynb" file
        with open(File.getPrefix(self.fileLocation)+'.ipynb', 'w') as f:
            # Get the jupyter cell list wrapper
            ipynb = Constant.getIpynb()
            # Wrap cells
            ipynb['cells'] = cCells
            # Overwrite old ipynb file
            json.dump(ipynb, f, indent=True)

    def _sanitize(self, Cells):
        '''
        Remove the extra attributes that thebe uses,
        that Jupyter does not
        '''
        for i, cell in enumerate(Cells):
            del Cells[i]['execution_count']
            del Cells[i]['changed']

    '''
    -------------------------------
    HTML conversion
    -------------------------------
    '''

    def _convert(self):
        '''
        Return a deep copy of cellList with code replaced with html-ized code
        '''

        # Deep copy cells so the original is not converted to HTML
        tempCells = copy.deepcopy(self.Cells)


        for cell in tempCells:
            # Preprocessing a code cell
            if cell['cell_type'] == 'code':
                # Highlight the Python syntax
                cell['source'] = \
                        [highlight(source, PythonLexer(), HtmlFormatter()) for source in cell['source']] 
                
                # Highlight standard output and error
                for output in cell['outputs']:
                    # Highlight the standard output
                    if output['output_type'] == 'execute_result':
                        output['data']['text/plain'] = \
                                [highlight(text, BashLexer(), HtmlFormatter()) \
                                for text in output['data']['text/plain']]
                    # Highlight the error
                    if output['output_type'] == 'error':
                        output['traceback'] = \
                                [highlight(text, BashLexer(), HtmlFormatter()) \
                                for text in output['traceback']]

            # Preprocessing a markdown cell
            if cell['cell_type'] == 'markdown':

                # Flatten the list so multi line latex delimiters 
                # are not separated by HTML elements as this would
                # break MathJax.
                cell['source'] = ''.join(cell['source'])

                # Remove any html that could interupt 
                # markdown conversion
                clean = re.compile('<.*>.*</.*>')
                cell['source'] = re.sub(clean, '', cell['source']) 

                # Convert the markdown

                # These arguments are used to let pypandoc
                # know to ignore latex
                pdoc_args = ['--standalone', '--mathjax'] 
                # Convert from markdown to HTML
                cell['source'] = \
                        pypandoc.convert_text(cell['source'], \
                        'html', format = 'md',\
                        extra_args = pdoc_args)

        return tempCells 
    def _isActive(self):
        if  self.executionThread:
            if self.executionThread.is_alive():
                return True
            else:
                return False
        else:
            return False



import time, sys, datetime, glob, re, sys, time, os, copy, logging, threading, queue
from copy import deepcopy
from io import StringIO
import thebe.core.constants as Constant
import thebe.core.output as output 
import thebe.core.logger as Logger
import thebe.core.update as Update
import thebe.core.html as Html

logger = Logger.getLogger('run.log', __name__)

def runNewCells(socketio, cellsToRun, globalScope, localScope, jc):
    '''
    Run each changed cell, returning the output.
    '''

    # Append cells containing updated output to this
    newCells = []

    # Toggle to true when the users code produces an
    # error so code execution can stop
    kill = False

    for cellCount, cell in enumerate(cellsToRun):

        # Run changed code if it is not markdown
        # and no prior cell has triggered an error
        if cell['changed']:
            socketio.emit('message', 'Running cell #%s...'%(cellCount))
            socketio.emit('loading', cellCount)
            if cell['cell_type'] != 'markdown' and not kill:
                logger.info('\n------------------------\nRunning cell #%s\nIn directory: %s\nWith code:\n%s\n-------------------------------'%(cellCount, os.getcwd(), cell['source']))

                # Execute the code from the cell, stream 
                # outputs using socketio, and return output
                outputs = jExecute(socketio, cell['source'], globalScope, localScope, jc)

                # Prevent subsequent execution of code
                # if in error was found
                if hasError(outputs):
                    kill = True

                # Add output data to cell
                cell['outputs'] = outputs

                # How does ipython do this?
                cell['changed'] = False
                cell['execution_count'] = cell['execution_count'] + 1
                logger.info('exe co: %s'%(cell['execution_count'],))

        # Append run cell the new cell list
        newCells.append(cell)

    # Stop the front end loading
    socketio.emit('stop loading')

    return newCells

def jExecute(socketio, code, globalScope, localScope, jc):
    '''
    '''

    code = ''.join(code)

    # Execute the code
    msg_id = jc.execute(code)

    # Collect the response payload
    # reply = jc.get_shell_msg(msg_id)

    # Get the execution status
    # When the execution state is "idle" it is complete
    io_msg_content = jc.get_iopub_msg(timeout=1)['content']

    # We're going to catch this here before we start polling
    if 'execution_state' in io_msg_content and io_msg_content['execution_state'] == 'idle':
        logger.debug('No output!')

    # Initialize the temp variable
    temp = {}

    # Initialize outputs
    outputs = []

    # Continue polling for execution to complete
    # which is indicated by having an execution state of "idle"
    while True:
        # Save the last message content. This will hold the solution.
        # The next one has the idle execution state indicating the execution
        # is complete, but not the stdout output
        temp = io_msg_content

        # Check the message for various possibilities
        if 'data' in temp: # Indicates completed operation
            if 'image/png' in temp['data']:
                plotData =  temp['data']['image/png']
                output = getPlotOutput(plotData)
                outputs.append(output)

                socketio.emit('output', output)

        if 'name' in temp and temp['name'] == "stdout": # indicates output
            # Create output for server use
            output = getStdOut(temp['text'])
            outputs.append(output)

            # Send HTML output for immediate front end use
            htmlOutput = deepcopy(output)
            htmlOutput['data']['text/plain'] = [Html.convertText(text, ttype = 'bash') for text in htmlOutput['data']['text/plain']]
            socketio.emit('output', htmlOutput)

        if 'evalue' in temp: # Indicates error

            # Create output for server use
            output = getErr(temp['evalue'])
            outputs.append(output)

            # Send HTML output for immediate front end use
            htmlOutput = deepcopy(output)
            htmlOutput['evalue'] = [Html.convertText(text, ttype = 'bash') for text in htmlOutput['evalue']]
            socketio.emit('output', htmlOutput)
            
            # If there is an error than it is pointless 
            # to keep on running code
            break

        # Poll the message
        try:
            logger.info('Retrieving message...')
            io_msg_content = jc.get_iopub_msg()['content']
            time.sleep(.1)

            if 'execution_state' in io_msg_content and io_msg_content['execution_state'] == 'idle':
                break

        except queue.Empty:
            break

    return outputs

def hasError(outputs):
    '''
    If an error cell exists in outputs
    return true
    '''
    for output in outputs:
        if 'evalue' in output:
            return True
    return False

def clearOutputs(cell):
    '''
    Replace list of outputs with an empty one.
    '''
    cell['outputs'] = []

def fillPlot(cell, plot):
    '''
    If an image exists in the plot variable, create and return a plot cell.
    '''
    if plot:
        output = Constant.getDisplayOutput()
        output['data']['image/png'] = plot
        cell['outputs'].append(output)
    return cell

def getPlotOutput(plot):
    '''
    '''
    output = Constant.getDisplayOutput()
    output['data']['image/png'] = plot
    return output

def getStdOut(stdOut):
    output = Constant.getExecuteOutput()
    output['data']['text/plain'] = stdOut.splitlines(True)
    return output

def getErr(err):
    output = Constant.getErrorOutput()
    output['evalue'] = err.splitlines(True)
    return output

def fillStdOut(cell, stdOut):
    '''
    If an output exists in the stdOut variable append new output to cell reference.
    '''
    if stdOut:
        output = Constant.getExecuteOutput()
        output['data']['text/plain'] = stdOut.splitlines(True)
        cell['outputs'].append(output)

def fillErr(cell, err):
    '''
    If an output exists in the err variable, create and return a err cell.
    '''
    if err:
        output = Constant.getErrorOutput()
        output['evalue'] = err.splitlines(True)
        cell['outputs'].append(output)

def runWithExec(socketio, cellCode, globalScope, localScope):
    '''
    runs one cell of code and return plotdata and std out/err
    '''
    #Remove magics
    newCellCode = [cell for cell in cellCode if cell[0] != '%']

    #Append the current working directory to path(not sure if this is necessary)
    sys.path.append(os.getcwd())
    
    #Save the old output location
    oldstdout = sys.stdout

    #Redirect system output, and initialize system error
    stdout = sys.stdout=StringIO()
    stderr = ''
    plotData = ''

    #Set cutoff var for output
    isRunning = queue.Queue()
    isRunning.put(True)
    #Create a thread to stream the output to the client during runtime.
    t = threading.Thread(target = Update.streamOutput, \
            args = (socketio, stdout, isRunning))  
    t.daemon = True
    t.start()

    #Run code and capture output, if there is an error stop running, and capture it.
    try:
        exec(''.join(newCellCode), globalScope, localScope)
        sys.stdout = oldstdout
        stdout = stdout.getvalue() 
        sys.path.pop()
        plotData = getPlotData(globalScope, localScope)

    except Exception as e:
#        globalScope = oldGlobalScope
#        localScope = oldLocalScope
        stderr = str(e)
        sys.stdout = oldstdout
        stdout = stdout.getvalue() 
        sys.path.pop()

    #Turn off streamOutput
    isRunning.put(False)

    return stdout, stderr, plotData

def getPlotData(globalScope, localScope):
    '''
    '''

    code=Constant.GetPlot
    redirected_output=sys.stdout=StringIO()
    redirected_error=sys.stderr=StringIO()
    stdout=''
    stderr=''
    sys.path.append(os.getcwd())
    try:
        exec(code, globalScope, localScope)
        stdout=redirected_output.getvalue().rstrip()
        stderr=''
    except Exception as e:
        stdout=redirected_output.getvalue()
        stderr=str(e)
    sys.path.pop()
    sys.stdout=sys.__stdout__
    sys.stderr=sys.__stderr__
    if stdout==Constant.EmptyGraph:
        stdout=''
    return stdout



import argparse

#Parse commandline argument
parser = argparse.ArgumentParser(description='Display python information live in browser.')
parser.add_argument('file', metavar='F', help='python file to run')
parser.add_argument('--port', '-p', default='5000', metavar='P', help='port number')
parser.add_argument('--debug', '-d', action='store_true', help='enable logging')
args = parser.parse_args()

def getDebug():
    return args.debug

def getFile():
    return args.file

def getPort():
    return int(args.port)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
        name='thebe',
        version='0.0.4.6',
        description='Automatically runs and displays python code in browser.',
        long_description=long_description,
        long_description_content_type="text/markdown",
        author_email='hairyhenry@gmail.com',
        url='https://github.com/hotsoupisgood/Satyrn',
        include_package_data=True,
        packages=find_packages(),
        install_requires=[
            'flask',
            'flask_socketio',
            'pygments',
            'dill',
            'pypandoc',
            'jupyter_client',
            ],
        entry_points={
            'console_scripts': [
                'thebe = thebe.thebe:main',
                ]
            },
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
        python_requires='>=3.7',
        )



import thebe.core.database as Database
import thebe.core.run as Run
import thebe.core.file as File
import thebe.core.html as Html
import thebe.core.data as data
import thebe.core.constants as Constants 
import thebe.core.logger as Logger
import os, time, json, threading, queue, copy

logger = Logger.getLogger('update.log', __name__)

def checkUpdate(socketio, fileLocation, connected=False, \
        isIpynb=False, GlobalScope=None, LocalScope=None, Cells=None, runAll=False, jc=None):
    '''
    '''
    
    '''
    If code is currently being executed,
    stop checkUpdate. Send some feedback to client.
    '''
    Cells, iGlobalScope, iLocalScope  = Database.getLedger(fileLocation)
    isActive = Database.getIsActive(fileLocation)

    if runAll:
        Cells = clearOutputs(Cells)

    #If the run all event is triggered
    if runAll:
        logger.info('Run All Has Been Triggered!')
        if isActive:
            socketio.emit('flash')
        else:
            Cells = []
            thread = update(socketio, fileLocation, GlobalScope, LocalScope, Cells, isIpynb, jc, runAll)
            time.sleep(.5)

    #If it's modified or if it's the first time it has run, update.
    elif isModified(fileLocation):
        if isActive:
            logger.info('flashing')
            socketio.emit('flash')
        else:
            thread = update(socketio, fileLocation, GlobalScope, LocalScope, Cells, isIpynb, jc)
            time.sleep(.5)

    elif connected==True:
        if not isActive:
            if not Cells:
                thread = update(socketio, fileLocation, GlobalScope, LocalScope, Cells, isIpynb, jc)
            else: 
                socketio.emit('show all', Html.convert(Cells))
        else:
            socketio.emit('show all', Html.convert(Cells))

    else:
        pass
    time.sleep(.5)

#Run code and send code and outputs to client
def update(socketio, fileLocation, GlobalScope, LocalScope, Cells, isIpynb, jc, runAll=False):
    isActive = Database.setIsActive(fileLocation)

    '''
    Get some variables from database
    '''

    '''
    Get target file
    '''
    fileContent=''
    with open(fileLocation, 'r') as file_content:
        fileContent=file_content.read()
    '''
    Look at the file to see if anything has changed
    in the data.
    Return an updated ipynb,
    with proper changed values.
    '''
    Cells = data.update(Cells, fileContent, runAll)
    socketio.emit('show all', Html.convert(Cells))

    '''
    Send a list of the cells that will run to the
    client so it can show what is loading.
    '''
    def runThread(Cells, GlobalScope, LocalScope):
        '''
        Run the newly changed cells and return their output.
        '''
        logger.info('------------------\nOutput before update\n-------------------------------\n%s'%([cell['outputs'] for cell in Cells],))
        Cells = Run.runNewCells(socketio, Cells, GlobalScope, LocalScope, jc)
        logger.info('------------------\nOutput after update\n-------------------------------\n%s'%([cell['outputs'] for cell in Cells],))
        executions = Database.getExecutions(fileLocation)
        executions += 1
        logger.info('The number of code executions is %d' % executions)
#        socketio.emit('show all', Html.convert(Cells))
        
        '''
        Update the database with the fresh code.
        And outputs.
        '''
        Database.setActive(fileLocation, False)
        Database.update(fileLocation, Cells, GlobalScope, LocalScope, executions)
        updateIpynb(fileLocation, Cells)
    t = threading.Thread(target = runThread, args = (Cells, GlobalScope, LocalScope))
    t.daemon = True
    t.start()
    return t

def updateIpynb(fileLocation, Cells):
    '''
    Write the new changes to the ipynb file.
    '''
    cCells = copy.deepcopy(Cells)

    # Remove extra attributes created by thebe.
    sanitize(cCells)
    

    # Save cells into a ".ipynb" file
    with open(File.getPrefix(fileLocation)+'.ipynb', 'w') as f:
        # Get the jupyter cell list wrapper
        ipynb = Constants.getIpynb()
        # Wrap cells
        ipynb['cells'] = cCells
        # Overwrite old ipynb file
        json.dump(ipynb, f, indent=True)

def sanitize(Cells):
    '''
    Remove the extra attributes that thebe uses,
    that Jupyter does not
    '''
    for i, cell in enumerate(Cells):
        del Cells[i]['execution_count']
        del Cells[i]['changed']

def isModified(fileLocation, x=.3):
    '''
    Return true if the target file has been modified in the past x amount of time
    '''

    lastModified=os.path.getmtime(fileLocation)
    timeSinceModified=int(time.time()-lastModified)

    if timeSinceModified<=x:
        return True
    else:
        return False

def streamOutput(socketio, stream, isRunning):
    '''
    Send new output to the client on the fly.

    Intended to run for the duration of one cell.
    Which is when the queue, isRunning, is false.
    '''
    oldStream = ''
    while list(isRunning.queue)[-1]:
        currentStream = stream.getvalue()
        if not currentStream == oldStream:
            socketio.emit('output', currentStream.split('\n'))
            oldStream = currentStream
        # Sleep the loop so it doesn't pollute the socket. The length is currently arbitrary.
        time.sleep(.2)

def clearOutputs(Cells):
    '''
    Clearly outputs and return the new cells
    '''
    for cell in Cells:
        cell['outputs'] = []
    return Cells
